#include "../../nestypes.h"
#include "../kmsnddev.h"
#include "../../format/audiosys.h"
#include "../../format/handler.h"
#include "../../format/nsf6502.h"
#include "logtable.h"
#include "m_nsf.h"
#include "s_n106.h"

#define NES_BASECYCLES (21477270)
#define CPS_SHIFT (20)
#define PHASE_SHIFT (16+2)
#define  N106SOUND_DATA 0x4800
#define  N106SOUND_ADDR 0xF800

typedef struct {
	Uint32 logvol;
	Uint32 cycles;
	Uint32 spd;
	Uint32 phase;
	Uint32 tlen;

	Uint8 update;
	Uint8 freql;
	Uint8 freqm;
	Uint8 freqh;
	Uint8 vreg;
	Uint8 tadr;
	Uint8 nazo;
	Uint8 mute;
} N106_WM;

typedef struct {
	Uint32 cps;
	Uint32 mastervolume;

	N106_WM ch[8];

	Uint8 addressauto;
	Uint8 address;
	Uint8 chinuse;

	Uint32 tone[0x100];	/* TONE DATA */
	Uint8 data[0x80];
} N106SOUND;

__inline static void UPDATE(N106_WM *chp)
{
	if (chp->update & 3)
	{
		Uint32 freq;
		freq  = ((int)chp->freql);
		freq += ((int)chp->freqm) << 8;
		freq += ((int)chp->freqh) << 16;
		chp->spd = freq & 0x3ffff;
	}
	if (chp->update & 2)
	{
		Uint32 tlen;
		tlen = (0x20 - (chp->freqh & 0x1c)) << PHASE_SHIFT;
		if (chp->tlen != tlen)
		{
			chp->tlen = tlen;
			chp->phase = 0;
		}
	}
	if (chp->update & 4)
	{
		chp->logvol = LinearToLog((chp->vreg & 0x0f) << 2);
	}
	chp->update = 0;
}

static Int32 __fastcall N106SoundRender(void* pNezPlay)
{
	N106SOUND *n106s = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->n106s;
	N106_WM *chp;
	Int32 accum = 0;
	for (chp = &n106s->ch[8 - n106s->chinuse]; chp < &n106s->ch[8]; chp++)
	{
		Uint32 cyclesspd = n106s->chinuse << CPS_SHIFT;
		if (chp->update) UPDATE(chp);
		chp->cycles += n106s->cps;
		while (chp->cycles >= cyclesspd)
		{
			chp->cycles -= cyclesspd;
			chp->phase += chp->spd;
		}
		while (chp->phase >= chp->tlen) chp->phase -= chp->tlen;
		if (chp->mute) continue;
		accum += LogToLinear(n106s->tone[((chp->phase >> PHASE_SHIFT) + chp->tadr) & 0xff] + chp->logvol + n106s->mastervolume, LOG_LIN_BITS - LIN_BITS - LIN_BITS - 10);
	}
	return accum;
}

const static NES_AUDIO_HANDLER s_n106_audio_handler[] = {
	{ 1, N106SoundRender, }, 
	{ 0, 0, }, 
};

static void __fastcall N106SoundVolume(void* pNezPlay, Uint volume)
{
	N106SOUND *n106s = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->n106s;
	n106s->mastervolume = (volume << (LOG_BITS - 8)) << 1;
}

const static NES_VOLUME_HANDLER s_n106_volume_handler[] = {
	{ N106SoundVolume, }, 
	{ 0, }, 
};

static void __fastcall N106SoundWriteAddr(void *pNezPlay, Uint address, Uint value)
{
	N106SOUND *n106s = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->n106s;
	n106s->address     = (Uint8)(value & 0x7f);
	n106s->addressauto = (value & 0x80) ? 1 : 0;
}

static void __fastcall N106SoundWriteData(void *pNezPlay, Uint address, Uint value)
{
	N106SOUND *n106s = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->n106s;
	n106s->data[n106s->address] = (Uint8)value;
	n106s->tone[n106s->address * 2]     = LinearToLog(((Int32)(value & 0xf) << 2) - 0x20);
	n106s->tone[n106s->address * 2 + 1] = LinearToLog(((Int32)(value >>  4) << 2) - 0x20);
	if (n106s->address >= 0x40)
	{
		N106_WM *chp = &n106s->ch[(n106s->address - 0x40) >> 3];
		switch (n106s->address & 7)
		{
			case 0:
				chp->update |= 1;
				chp->freql = (Uint8)value;
				break;
			case 2:
				chp->update |= 1;
				chp->freqm = (Uint8)value;
				break;
			case 4:
				chp->update |= 2;
				chp->freqh = (Uint8)value;
				break;
			case 6:
				chp->tadr = (Uint8)(value & 0xff);
				break;
			case 7:
				chp->update |= 4;
				chp->vreg = (Uint8)value;
				chp->nazo = (Uint8)((value >> 4) & 0x07);
				if (chp == &n106s->ch[7])
					n106s->chinuse = 1 + chp->nazo;
				break;
		}
	}
	if (n106s->addressauto)
	{
		n106s->address = (n106s->address + 1) & 0x7f;
	}
}

static NES_WRITE_HANDLER s_n106_write_handler[] =
{
	{ N106SOUND_DATA, N106SOUND_DATA, N106SoundWriteData, },
	{ N106SOUND_ADDR, N106SOUND_ADDR, N106SoundWriteAddr, },
	{ 0,              0,              0, },
};

static Uint __fastcall N106SoundReadData(void *pNezPlay, Uint address)
{
	N106SOUND *n106s = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->n106s;
	Uint ret = n106s->data[n106s->address];
	if (n106s->addressauto)
	{
		n106s->address = (n106s->address + 1) & 0x7f;
	}
	return ret;
}

static NES_READ_HANDLER s_n106_read_handler[] =
{
	{ N106SOUND_DATA, N106SOUND_DATA, N106SoundReadData, },
	{ 0,              0,              0, },
};

static Uint32 DivFix(Uint32 p1, Uint32 p2, Uint32 fix)
{
	Uint32 ret;
	ret = p1 / p2;
	p1  = p1 % p2;/* p1 = p1 - p2 * ret; */
	while (fix--)
	{
		p1 += p1;
		ret += ret;
		if (p1 >= p2)
		{
			p1 -= p2;
			ret++;
		}
	}
	return ret;
}

static void __fastcall N106SoundReset(void* pNezPlay)
{
	N106SOUND *n106s = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->n106s;
	int i;
	XMEMSET(n106s, 0, sizeof(N106SOUND));
	for (i = 0; i < 8; i++)
	{
		n106s->ch[i].tlen = 0x10 << PHASE_SHIFT;
		n106s->ch[i].logvol = LinearToLog(0);
	}
	n106s->addressauto = 1;
	n106s->chinuse = 8;
	n106s->cps = DivFix(NES_BASECYCLES, 45 * NESAudioFrequencyGet(pNezPlay), CPS_SHIFT);
}

const static NES_RESET_HANDLER s_n106_reset_handler[] = {
	{ NES_RESET_SYS_NOMAL, N106SoundReset, }, 
	{ 0,                   0, }, 
};


static void __fastcall N106SoundTerm(void* pNezPlay)
{
	N106SOUND *n106s = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->n106s;
	if (n106s)
		XFREE(n106s);
}

const static NES_TERMINATE_HANDLER s_n106_terminate_handler[] = {
	{ N106SoundTerm, }, 
	{ 0, }, 
};

void N106SoundInstall(NEZ_PLAY *pNezPlay)
{
	N106SOUND *n106s;
	n106s = XMALLOC(sizeof(N106SOUND));
	if (!n106s) return;
	XMEMSET(n106s, 0, sizeof(N106SOUND));
	((NSFNSF*)pNezPlay->nsf)->n106s = n106s;

	LogTableInitialize();
	NESAudioHandlerInstall(pNezPlay, s_n106_audio_handler);
	NESVolumeHandlerInstall(pNezPlay, s_n106_volume_handler);
	NESTerminateHandlerInstall(&pNezPlay->nth, s_n106_terminate_handler);
	NESReadHandlerInstall(pNezPlay, s_n106_read_handler);
	NESWriteHandlerInstall(pNezPlay, s_n106_write_handler);
	NESResetHandlerInstall(pNezPlay->nrh, s_n106_reset_handler);
}
